// winstack.cpp : implementation file
//

#include "stdafx.h"
#include "winstack.h"
#include "winutil.h"
#include "resource.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// page flags for internal use only
#define SWF_INT_CAPTIONPRESSED   0x8000

// timer id for delayed show panning cursor
#define TIMER_PANCURSOR_ID       100
#define TIMER_PANCURSOR_DELAY    300

/////////////////////////////////////////////////////////////////////////////
// CStackWnd::CPageInfo

CStackWnd::CPageInfo::CPageInfo()
{
   m_pDlg = NULL;
   m_lpszTemplate = NULL;
   m_dwFlags = 0;
   m_lpszText = NULL;
   m_pNext = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// CStackWnd

IMPLEMENT_DYNCREATE(CStackWnd, CWnd)

CStackWnd::CStackWnd()
{
   m_pPageInfo = NULL;
   m_pPageInfoTemp = NULL;
   m_nPageCount = 0;
   m_cyCaption = 18;
   m_cyGap = 2;
   m_xOrigin = 0;
   m_yOrigin = 0;
   m_bPanning = FALSE;
   m_hCurPan = NULL;
   m_bCurPan = FALSE;
   m_hCurDefault = ::LoadCursor(NULL, IDC_ARROW);
   m_sizeMin.cx = m_sizeMin.cy = 100;
}

CStackWnd::~CStackWnd()
{
   if (m_pPageInfo)
      ClearPageInfo();

   m_dottedBrush.DeleteObject();
}

BOOL CStackWnd::PreCreateWindow(CREATESTRUCT& cs) 
{
	if (!CWnd::PreCreateWindow(cs))
      return FALSE;

   cs.lpszClass = AfxRegisterWndClass(CS_DBLCLKS, NULL, NULL, NULL);
   cs.style |= WS_CLIPCHILDREN;
   return TRUE;
}

BOOL CStackWnd::PreTranslateMessage(MSG* pMsg) 
{
	ASSERT(pMsg != NULL);
   ASSERT_VALID(this);
   ASSERT(m_hWnd != NULL);

	if (CWnd::PreTranslateMessage(pMsg))
      return TRUE;

   // don't translate dialog messages when 
   // application is in help mode
   CFrameWnd* pFrameWnd = GetTopLevelFrame();
   if (pFrameWnd != NULL && pFrameWnd->m_bHelpMode)
      return FALSE;

   // ensure the dialog messages will not
   // eat frame accelerators
   pFrameWnd = GetParentFrame();
   while (pFrameWnd != NULL)
   {
      if (pFrameWnd->PreTranslateMessage(pMsg))
         return TRUE;
      pFrameWnd = pFrameWnd->GetParentFrame();
   }

   return PreTranslateInput(pMsg);
}

BOOL CStackWnd::InsertPage(CDialog* pDlg, LPCTSTR lpszTemplate, 
   LPCTSTR lpszCaption, DWORD dwFlags, int nPage)
{
   ASSERT_VALID(this);
   ASSERT(pDlg != NULL);
   ASSERT(pDlg->m_hWnd == NULL);

   // load dialog resource
   if (!WU_CheckDlgTemplate(lpszTemplate, TRUE))
   {
      ASSERT(FALSE);
      PostNcDestroy();
      return FALSE;
   }

   // create a modeless child dialog
   BOOL bSuccess = TRUE;
   TRY
   {
      // create dialog box
      if (!pDlg->Create(lpszTemplate, this))
         AfxThrowUserException();
      
      // get dialog box size
      CRect rectDlg;
      pDlg->GetWindowRect(&rectDlg);

      // setup page info
      CPageInfo* pInfo = new CPageInfo;
      if (pInfo == NULL)
         AfxThrowMemoryException();

      pInfo->m_pDlg = pDlg;
      pInfo->m_lpszTemplate = lpszTemplate;
      pInfo->m_sizeDlg = CSize(rectDlg.Width(), rectDlg.Height());
      pInfo->m_lpszText = lpszCaption;
      pInfo->m_dwFlags = dwFlags & 0x00FF;

      // set dialog ID
      nPage = InsertPageInfo(pInfo, nPage);
      pDlg->SetDlgCtrlID(AFX_IDW_PANE_FIRST+nPage+1);

      // adjust maximum width
      if (rectDlg.Width() > m_sizeStack.cx)
         m_sizeStack.cx = rectDlg.Width();

      // save page count
      ++m_nPageCount;

      // recalculate window layout
      RecalcLayout();
      RedrawWindow();
   }
   CATCH_ALL(e)
   {
      TRACE0("Failed to create dialog in CStackWnd::InsertPage()\n");
      bSuccess = FALSE;
   }
   END_CATCH_ALL;

   return bSuccess;
}

BOOL CStackWnd::InsertPage(CDialog* pDlg, UINT nIDTemplate, 
   LPCTSTR lpszCaption, DWORD dwFlags, int nPage)
{
   return InsertPage(pDlg, MAKEINTRESOURCE(nIDTemplate),
      lpszCaption, dwFlags, nPage);
}

int CStackWnd::InsertPageInfo(CPageInfo* pInfo, int nPage)
{
   ASSERT(pInfo != NULL);

   if (m_pPageInfo != NULL)
   {
      if (nPage == 0)
      {
         pInfo->m_pNext = m_pPageInfo;
         m_pPageInfo = pInfo;
      }
      else
      {
         int n = 0;
         CPageInfo* ptr = m_pPageInfo;
         while (ptr->m_pNext)
         {
            if (nPage == n+1)
            {
               n++;
               break;
            }
            n++;
            ptr = ptr->m_pNext;
         }

         ASSERT(ptr != NULL);
         if (ptr->m_pNext == NULL)
            n++;

         pInfo->m_pNext = ptr->m_pNext;
         ptr->m_pNext = pInfo;
         return n;
      }
   }
   else
      m_pPageInfo = pInfo;

   return 0;
}

void CStackWnd::ClearPageInfo()
{
   CPageInfo* ptr = m_pPageInfo, *ptrNext;
   while (ptr)
   {
      ptrNext = ptr->m_pNext;
      delete ptr;
      ptr = ptrNext;
   }

   m_pPageInfo = NULL;
}

BOOL CStackWnd::IsPage(HWND hWnd)
{
   if (!::IsChild(m_hWnd, hWnd))
      return FALSE;

   CPageInfo* pInfo = m_pPageInfo;
   while (pInfo != NULL)
   {
      if (pInfo->m_pDlg->m_hWnd == hWnd)
         return TRUE;
      pInfo = pInfo->m_pNext;
   }

   return FALSE;
}

CStackWnd::CPageInfo* CStackWnd::CaptionHitTest(POINT pt)
{
   CRect rect;
   CPageInfo* pInfo = m_pPageInfo;
   while (pInfo != NULL)
   {
      if (pInfo->m_dwFlags & SWF_VISIBLE)
      {
         CalcCaptionRect(pInfo, &rect);
         if (rect.PtInRect(pt))
            return pInfo;
      }

      pInfo = pInfo->m_pNext;
   }

   return NULL;
}  

void CStackWnd::CalcCaptionRect(CPageInfo* pInfo, LPRECT lpRect)
{
   ASSERT(pInfo != NULL);
   ASSERT(pInfo->m_dwFlags & SWF_VISIBLE);
   ASSERT(lpRect != NULL);

   *lpRect = pInfo->m_rectPage;
   lpRect->bottom = lpRect->top+m_cyCaption;
   OffsetRect(lpRect, m_xOrigin, m_yOrigin);
}

BOOL CStackWnd::IsPanEnabled(BOOL bPanX, BOOL bPanY)
{
   CRect rectClient;
   GetClientRect(&rectClient);
   return ((bPanX && m_sizeStack.cx > rectClient.Width()) ||
           (bPanY && m_sizeStack.cy > rectClient.Height())) ? TRUE : FALSE;
}

void CStackWnd::RecalcLayout(BOOL bRepositionPages)
{
   ASSERT_VALID(this);
   ASSERT(::IsWindow(m_hWnd));

   // get client rectangle
   CRect rectClient;
   GetClientRect(&rectClient);

   CDialog* pDlgPage;
   CRect    rectPage;
   DWORD    dwSWPFlags;
   int      yc = 0;
   HDWP     hDWP;

   if (m_nPageCount > 0 && bRepositionPages)
   {
      hDWP = BeginDeferWindowPos(m_nPageCount);
      if (hDWP == NULL)
      {
         TRACE0("Low system resources in CStackWnd::RecalcLayout()\n");
         AfxThrowMemoryException();
      }
   }

   CPageInfo* pInfo = m_pPageInfo;
   while (pInfo != NULL)
   {
#ifdef _DEBUG
      ASSERT(pInfo->m_pDlg != NULL);
      ASSERT(pInfo->m_pDlg->m_hWnd != NULL);
#endif // _DEBUG
      pDlgPage = pInfo->m_pDlg;
      if (pInfo->m_dwFlags & SWF_VISIBLE)
      {
         // flags for SetWindowPos
         dwSWPFlags = SWP_NOZORDER | SWP_NOREDRAW | SWP_NOSIZE;

         // calculate page rectangle
         rectPage.SetRect(0, yc, m_sizeStack.cx, yc+m_cyCaption+m_cyGap);
         if (!(pInfo->m_dwFlags & SWF_CONTRACTED))
         {
            rectPage.bottom += pInfo->m_sizeDlg.cy;
            if (!pDlgPage->IsWindowVisible())
               dwSWPFlags |= SWP_SHOWWINDOW;
         }
         else
            dwSWPFlags |= SWP_HIDEWINDOW;

         yc += rectPage.Height();
         pInfo->m_rectPage = rectPage;

         // calculate rectangle for child dialog
         if (!(pInfo->m_dwFlags & SWF_CONTRACTED))
         {
            rectPage.top += m_cyCaption+m_cyGap;
            rectPage.right = rectPage.left+pInfo->m_sizeDlg.cx;
         }

         // reposition child dialog
         if (bRepositionPages)
            DeferWindowPos(hDWP, pDlgPage->m_hWnd, NULL, rectPage.left+m_xOrigin,
               rectPage.top+m_yOrigin, rectPage.Width(), 
               rectPage.Height(), dwSWPFlags);
      }
      else
      {
         // hide window
         if (bRepositionPages)
            if (pDlgPage->IsWindowVisible())
               DeferWindowPos(hDWP, pDlgPage->m_hWnd, NULL, 0, 0, 0, 0,
                  SWP_NOZORDER | SWP_HIDEWINDOW);
      }

      // get next page
      pInfo = pInfo->m_pNext;
   }

   // reposition child all window at once
   if (m_nPageCount > 0 && bRepositionPages)
      EndDeferWindowPos(hDWP);

   // save stack height for further operations
   m_sizeStack.cy = yc;
}

void CStackWnd::SetPageState(CPageInfo* pInfo, BOOL bShow, BOOL bContract, BOOL bEnsureVisible)
{
   ASSERT_VALID(this);
   ASSERT(m_hWnd != NULL);
   ASSERT(pInfo != NULL);
   ASSERT_VALID(pInfo->m_pDlg);
   ASSERT(pInfo->m_pDlg->m_hWnd != NULL);

   BOOL  bStateChanged = FALSE, bRemoveFocus = FALSE;
   CSize dist;
   HWND  hWndFocus;

   // change page state flags
   if (bShow)
   {
      if (!(pInfo->m_dwFlags & SWF_VISIBLE))
      {
         pInfo->m_dwFlags |= SWF_VISIBLE;
         bStateChanged = TRUE;
      }
   }
   else
   {
      if (pInfo->m_dwFlags & SWF_VISIBLE)
      {
         pInfo->m_dwFlags &= ~SWF_VISIBLE;
         bEnsureVisible = FALSE;
         bStateChanged = TRUE;

         hWndFocus = ::GetFocus();
         if (::IsWindow(hWndFocus) && ::IsChild(pInfo->m_pDlg->m_hWnd, hWndFocus))
            bRemoveFocus = TRUE;
      }
   }

   if (bContract)
   {
      if (!(pInfo->m_dwFlags & SWF_CONTRACTED))
      {
         pInfo->m_dwFlags |= SWF_CONTRACTED;
         bEnsureVisible = FALSE;
         bStateChanged = TRUE;

         hWndFocus = ::GetFocus();
         if (::IsWindow(hWndFocus) && ::IsChild(pInfo->m_pDlg->m_hWnd, hWndFocus))
            bRemoveFocus = TRUE;
      }
   }
   else
   {
      if (pInfo->m_dwFlags & SWF_CONTRACTED)
      {
         pInfo->m_dwFlags &= ~SWF_CONTRACTED;
         bStateChanged = TRUE;
      }
   }

   // state has changed - redraw window
   if (bStateChanged)
   {
      RecalcLayout(FALSE);

      dist.cx = 0;
      dist.cy = 0;

      if (bEnsureVisible)
      {
         if (pInfo->m_rectPage.right+m_xOrigin > m_sizeClient.cx)
         {
            dist.cx = m_sizeClient.cx-pInfo->m_rectPage.right-m_xOrigin;
            if (pInfo->m_rectPage.left+m_xOrigin+dist.cx < 0)
               dist.cx = -m_xOrigin-pInfo->m_rectPage.left;
         }

         if (pInfo->m_rectPage.bottom+m_yOrigin > m_sizeClient.cy)
         {
            dist.cy = m_sizeClient.cy-pInfo->m_rectPage.bottom-m_yOrigin;
            if (pInfo->m_rectPage.top+m_yOrigin+dist.cy < 0)
               dist.cy = -m_yOrigin-pInfo->m_rectPage.top;
         }
      }

      if (m_sizeStack.cx+m_xOrigin < m_sizeClient.cx)
      {
         dist.cx = m_sizeClient.cx-m_sizeStack.cx-m_xOrigin;
         if (m_xOrigin+dist.cx > 0)
            dist.cx = -m_xOrigin;
      }

      if (m_sizeStack.cy+m_yOrigin < m_sizeClient.cy)
      {
         dist.cy = m_sizeClient.cy-m_sizeStack.cy-m_yOrigin;
         if (m_yOrigin+dist.cy > 0)
            dist.cy = -m_yOrigin;
      }

      m_xOrigin += dist.cx;
      m_yOrigin += dist.cy;

      RecalcLayout();
      RedrawWindow(NULL, NULL, RDW_INVALIDATE | RDW_ERASE | 
         RDW_ALLCHILDREN | RDW_UPDATENOW);

      if (bRemoveFocus)
      {
         CFrameWnd* pFrameWnd = GetParentFrame();
         if (pFrameWnd != NULL)
            pFrameWnd->SetFocus();
      }
   }
}

void CStackWnd::DrawPageCaption(CDC* pDC, LPCTSTR lpszText, UINT nState, LPRECT lpRect)
{
   ASSERT_VALID(this);

   // is caption button pressed? expanded?
   BOOL bPressed = (nState & SWF_INT_CAPTIONPRESSED) ?
      TRUE : FALSE;
   BOOL bExpanded = (nState & SWF_CONTRACTED) ?
      FALSE : TRUE;

   // draw border
   COLORREF clrBorder[2];
   clrBorder[0] = GetSysColor(COLOR_3DHIGHLIGHT);
   clrBorder[1] = GetSysColor(COLOR_3DSHADOW);
   if (bPressed || bExpanded)
      pDC->Draw3dRect(lpRect, clrBorder[1], clrBorder[0]);
   else
      pDC->Draw3dRect(lpRect, clrBorder[0], clrBorder[1]);

   InflateRect(lpRect, -1, -1);

   if (bExpanded && !bPressed)
   {
      pDC->SetTextColor(GetSysColor(COLOR_3DHIGHLIGHT));
      pDC->SetBkColor(GetSysColor(COLOR_3DFACE));
      pDC->FillRect(lpRect, &m_dottedBrush);
   }
   else
      pDC->FillSolidRect(lpRect, GetSysColor(COLOR_3DFACE));

   if (bPressed || bExpanded)
      OffsetRect(lpRect, 1, 1);

   // draw text
   if (lpszText)
   {
      CFont font;
      LOGFONT lf;
      WU_GetDefaultGUIFont(8, &lf);
//    lf.lfWeight = FW_BOLD;
      font.CreateFontIndirect(&lf);
      pDC->SelectObject(font);
      pDC->SetTextColor(GetSysColor(COLOR_BTNTEXT));
      pDC->SetBkMode(TRANSPARENT);
      pDC->DrawText(lpszText, lpRect, DT_CENTER | DT_VCENTER);
   }
}

BEGIN_MESSAGE_MAP(CStackWnd, CWnd)
	//{{AFX_MSG_MAP(CStackWnd)
	ON_WM_CREATE()
	ON_WM_PAINT()
	ON_WM_SIZE()
	ON_WM_SETCURSOR()
	ON_WM_MOUSEMOVE()
	ON_WM_LBUTTONUP()
	ON_WM_LBUTTONDOWN()
	ON_WM_WINDOWPOSCHANGED()
	ON_WM_CANCELMODE()
	ON_WM_TIMER()
	ON_WM_LBUTTONDBLCLK()
	//}}AFX_MSG_MAP
   ON_MESSAGE(LWM_CALCPANELSIZE, OnCalcPanelSize)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CStackWnd message handlers

void CStackWnd::CreateDottedBrush()
{
   m_dottedBrush.DeleteObject();
   WORD wPatternBitmap[8];
   WORD wPattern = 0xAA55;
   for (int i = 0; i < 8; i++)
   {
      wPatternBitmap[i] = wPattern;
      wPattern = MAKEWORD(HIBYTE(wPattern), LOBYTE(wPattern));
   }

   CBitmap bmpPattern;
   bmpPattern.CreateBitmap(8, 8, 1, 1, wPatternBitmap);
   m_dottedBrush.CreatePatternBrush(&bmpPattern);
   bmpPattern.DeleteObject();
}

int CStackWnd::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CWnd::OnCreate(lpCreateStruct) == -1)
		return -1;

   CreateDottedBrush();

   // load panning hand cursor
   m_hCurPan = AfxGetApp()->LoadCursor(IDC_PAN);
   if (m_hCurPan == NULL)
      m_hCurPan = ::LoadCursor(NULL, IDC_SIZEALL);

	return 0;
}

LRESULT CStackWnd::OnCalcPanelSize(WPARAM wParam, LPARAM lParam)
{
   SIZE* pSize = (SIZE*)lParam;
   ASSERT(pSize != NULL);
   pSize->cx = (m_sizeStack.cx > 0) ? 
      m_sizeStack.cx : m_sizeMin.cx;
   pSize->cy = m_sizeMin.cy;
   return TRUE;
}

void CStackWnd::OnPaint()
{
	CPaintDC dc(this);

   // system defined dialog background color
   COLORREF clr3DFace = GetSysColor(COLOR_3DFACE);
   CRect rect, rectClient, rectCaption;
   int bottom;

   GetClientRect(&rectClient);
   bottom = 0;

   CPageInfo* pInfo = m_pPageInfo;
   while (pInfo != NULL)
   {
      // get page rectangle
      rect = pInfo->m_rectPage;
      rect.OffsetRect(m_xOrigin, m_yOrigin);
      bottom = rect.bottom;

      if (rect.top > rectClient.bottom)
         break;

      // draw caption
      rectCaption = rect;
      rectCaption.bottom = rect.top+m_cyCaption;
      DrawPageCaption(&dc, pInfo->m_lpszText, 
         pInfo->m_dwFlags, &rectCaption);

      // fill gap
      rect.top += m_cyCaption;
      rect.bottom = rect.top + m_cyGap;
      dc.FillSolidRect(&rect, clr3DFace);

      // fill right hole
      if (!(pInfo->m_dwFlags & SWF_CONTRACTED))
      {
         rect.top += m_cyGap;
         rect.left += pInfo->m_sizeDlg.cx;
         rect.bottom = rect.top+pInfo->m_sizeDlg.cy;
         if (rect.Width() > 0)
            dc.FillSolidRect(&rect, clr3DFace);
      }

      // get next page
      pInfo = pInfo->m_pNext;
   }

   // fill bottom hole
   if (bottom < rectClient.bottom)
   {
      dc.FillSolidRect(0, bottom, rectClient.right, 
         rectClient.bottom, clr3DFace);
   }
}

void CStackWnd::OnSize(UINT nType, int cx, int cy) 
{
	CWnd::OnSize(nType, cx, cy);

   if (nType != SIZE_MINIMIZED && cx > 0 && cy > 0)
   {
      CSize distScroll(0, 0);
      CRect rectClient, rectUX, rectUY;

      rectClient.SetRect(0, 0, cx, cy);
      rectUX = rectClient;
      rectUY = rectClient;

      // calculate update region (right)
      if (cx > m_sizeClient.cx)
         rectUX.right -= cx-m_sizeClient.cx;

      // calculate scroll distance (X)
      if (m_sizeStack.cx+m_xOrigin < cx)
      {
         distScroll.cx = cx-m_sizeStack.cx-m_xOrigin;
         if (m_xOrigin+distScroll.cx > 0)
            distScroll.cx = -m_xOrigin;

         rectUX.left += distScroll.cx;
      }

      // calculate update region (bottom)
      if (cy > m_sizeClient.cy)
         rectUY.bottom -= cy-m_sizeClient.cy;

      // calculate scroll distance (Y)
      if (m_sizeStack.cy+m_yOrigin < cy)
      {
         distScroll.cy = cy-m_sizeStack.cy-m_yOrigin;
         if (m_yOrigin+distScroll.cy > 0)
            distScroll.cy = -m_yOrigin;

         rectUY.top += distScroll.cy;
      }

      // change offsets
      m_xOrigin += distScroll.cx;
      m_yOrigin += distScroll.cy;

      CRgn rgnA, rgnB, rgnC;
      rgnA.CreateRectRgnIndirect(&rectClient);
      rgnB.CreateRectRgnIndirect(&rectUX);
      rgnC.CreateRectRgn(0, 0, 0, 0);
#ifndef _DEGUG
      rgnC.CombineRgn(&rgnA, &rgnB, RGN_XOR);
#else // !_DEBUG
      int res = rgnC.CombineRgn(&rgnA, &rgnB, RGN_XOR);
      ASSERT(res != ERROR);
      if (res == NULLREGION)
         TRACE0("Info: region is empty in CStackWnd::OnSize()\n");
#endif // _DEBUG

      if (distScroll.cx > 0)
         ScrollWindowEx(distScroll.cx, 0, NULL, NULL, &rgnC, NULL,
            SW_ERASE | SW_INVALIDATE | SW_SCROLLCHILDREN);
      else
         RedrawWindow(NULL, &rgnC, RDW_INVALIDATE | RDW_ERASE | 
            RDW_ALLCHILDREN | RDW_UPDATENOW);

      rgnA.DeleteObject();
      rgnB.DeleteObject();
      rgnC.DeleteObject();

      rgnA.CreateRectRgnIndirect(&rectClient);
      rgnB.CreateRectRgnIndirect(&rectUY);
      rgnC.CreateRectRgn(0, 0, 0, 0);
#ifndef _DEGUG
      rgnC.CombineRgn(&rgnA, &rgnB, RGN_XOR);
#else // !_DEBUG
      int res = rgnC.CombineRgn(&rgnA, &rgnB, RGN_XOR);
      ASSERT(res != ERROR);
      if (res == NULLREGION)
         TRACE0("Info: region is empty in CStackWnd::OnSize()\n");
#endif // _DEBUG

      if (distScroll.cy > 0)
         ScrollWindowEx(0, distScroll.cy, NULL, NULL, &rgnC, NULL,
            SW_ERASE | SW_INVALIDATE | SW_SCROLLCHILDREN);
      else
         RedrawWindow(NULL, &rgnC, RDW_INVALIDATE | RDW_ERASE | 
            RDW_ALLCHILDREN | RDW_UPDATENOW);

      rgnA.DeleteObject();
      rgnB.DeleteObject();
      rgnC.DeleteObject();

      // recalculate window layout
      RecalcLayout(FALSE);

      // save new size of client window
      m_sizeClient.cx = cx;
      m_sizeClient.cy = cy;
   }
}

void CStackWnd::OnCancelMode() 
{
	CWnd::OnCancelMode();
	
   // cancel panning
   if (m_bPanning)
      EndPanning();
   else
   {
      if (m_pPageInfoTemp)
      {
         BOOL bRedraw;
         CRect rect;

         bRedraw = ((m_pPageInfoTemp->m_dwFlags & SWF_INT_CAPTIONPRESSED) != 0);
         m_pPageInfoTemp->m_dwFlags &= ~SWF_INT_CAPTIONPRESSED;

         if (bRedraw)
         {
            CalcCaptionRect(m_pPageInfoTemp, &rect);
            CDC* pDC = GetDC();
            DrawPageCaption(pDC, m_pPageInfoTemp->m_lpszText, 
               m_pPageInfoTemp->m_dwFlags, &rect);
            ReleaseDC(pDC);
         }

         ReleaseCapture();
         m_pPageInfoTemp = NULL;
      }
   }
}

/////////////////////////////////////////////////////////////////////////////
// Interaction

BOOL CStackWnd::BeginPanning(POINT pt)
{
   SetCursor(m_hCurPan);
   m_ptSave = pt;
   SetCapture();
   return TRUE;
}

void CStackWnd::ContinuePanning(POINT pt)
{
   // get client rectangle
   CRect rectClient, rectUX, rectUY;
   GetClientRect(&rectClient);

   // calculate offsets
   CSize distPan(0, 0);
   rectUX = rectClient;
   rectUY = rectClient;

   if (IsPanEnabled(TRUE, FALSE))
   {
      distPan.cx = pt.x-m_ptSave.x;

      if (distPan.cx < 0)
      {
         if (distPan.cx+m_xOrigin+m_sizeStack.cx < rectClient.right)
            distPan.cx = rectClient.right-m_xOrigin-m_sizeStack.cx;
         rectUX.left = rectUX.right+distPan.cx;
      }
      else
      {
         if (distPan.cx+m_xOrigin > 0)
            distPan.cx = -m_xOrigin;
         rectUX.right = rectUX.left+distPan.cx;
      }
   }

   if (IsPanEnabled(FALSE, TRUE))
   {
      distPan.cy = pt.y-m_ptSave.y;

      if (distPan.cy < 0)
      {
         if (distPan.cy+m_yOrigin+m_sizeStack.cy < rectClient.bottom)
            distPan.cy = rectClient.bottom-m_yOrigin-m_sizeStack.cy;
         rectUY.top = rectUY.bottom+distPan.cy;
      }
      else
      {
         if (distPan.cy+m_yOrigin > 0)
            distPan.cy = -m_yOrigin;
         rectUY.bottom = rectUY.top+distPan.cy;
      }
   }
   
   if (distPan.cx != 0 || distPan.cy != 0)
   {
      m_xOrigin += distPan.cx;
      m_yOrigin += distPan.cy;
   
      // scroll window
      if (distPan.cx != 0)
         ScrollWindowEx(distPan.cx, 0, NULL, NULL, NULL, &rectUX, 
            SW_ERASE | SW_INVALIDATE | SW_SCROLLCHILDREN);
      if (distPan.cy != 0)
         ScrollWindowEx(0, distPan.cy, NULL, NULL, NULL, &rectUY, 
            SW_ERASE | SW_INVALIDATE | SW_SCROLLCHILDREN);

      // recalculate layout (don't reposition pages)
      RecalcLayout(FALSE);
   }

   // save current position
   m_ptSave = pt;
}

void CStackWnd::EndPanning()
{
   ReleaseCapture();
   m_bCurPan = TRUE;
   m_bPanning = FALSE;
}

BOOL CStackWnd::DelayedShowPanCursor(BOOL bShow)
{
   if (bShow)
   {
      KillTimer(TIMER_PANCURSOR_ID);
      SetCursor(m_hCurPan);
      return TRUE;
   }
   else
   {
      SetCursor(m_hCurDefault);
      GetCursorPos(&m_ptSave);
      SetTimer(TIMER_PANCURSOR_ID, TIMER_PANCURSOR_DELAY, NULL);
   }

   return TRUE;
}

BOOL CStackWnd::BeginCaptionPress(POINT point)
{
   if (m_pPageInfoTemp)
      return FALSE;

   m_pPageInfoTemp = CaptionHitTest(point);
   if (m_pPageInfoTemp)
   {
      CDC* pDC = GetDC();
      CRect rect;
      CalcCaptionRect(m_pPageInfoTemp, &rect);
      m_pPageInfoTemp->m_dwFlags |= SWF_INT_CAPTIONPRESSED;
      DrawPageCaption(pDC, m_pPageInfoTemp->m_lpszText, 
         m_pPageInfoTemp->m_dwFlags, &rect);
      ReleaseDC(pDC);
      SetCapture();
      return TRUE;
   }

   return FALSE;
}

BOOL CStackWnd::OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message) 
{
   BOOL   bBeginPan = FALSE;
   CPoint pt;

   // get cursor position
   GetCursorPos(&pt);

   // get a window where the event occurs in
   // client coordinates
   HWND hWnd = ::WindowFromPoint(pt);
   ScreenToClient(&pt);

   // this is a stack window -- test caption click
   if (hWnd == m_hWnd)
      bBeginPan = (CaptionHitTest(pt) == NULL);
   else
   {
      // test click on child dialog but not on
      // nonstatic control of the child dialog
      if (IsPage(hWnd))
         bBeginPan = TRUE;
   }

   // begin panning
   if (bBeginPan && IsPanEnabled())
   {
      DelayedShowPanCursor(m_bCurPan);

      if (message == WM_LBUTTONDOWN)
         m_bPanning = BeginPanning(pt);
      return TRUE;
   }
   else
   {
      if (hWnd == m_hWnd)
      {
         SetCursor(m_hCurDefault);
         m_bCurPan = FALSE;
         return TRUE;
      }
   }

   // default cursor processing
   m_bCurPan = FALSE;
	return CWnd::OnSetCursor(pWnd, nHitTest, message);
}

void CStackWnd::OnTimer(UINT nIDEvent) 
{
   // delayed show panning cursor
   if (nIDEvent == TIMER_PANCURSOR_ID)
   {
      CPoint pt;
      GetCursorPos(&pt);
      if (pt == m_ptSave)
      {
         DelayedShowPanCursor(TRUE);
         m_bCurPan = TRUE;
      }
   }

	CWnd::OnTimer(nIDEvent);
}

void CStackWnd::OnMouseMove(UINT nFlags, CPoint point) 
{
	if (m_bPanning)
      ContinuePanning(point);
   else
   {
      if (m_pPageInfoTemp)
      {
         BOOL bRedraw;
         CRect rect;

         if (CaptionHitTest(point) == m_pPageInfoTemp)
         {
            bRedraw = ((m_pPageInfoTemp->m_dwFlags & SWF_INT_CAPTIONPRESSED) == 0);
            m_pPageInfoTemp->m_dwFlags |= SWF_INT_CAPTIONPRESSED;
         }
         else
         {
            bRedraw = ((m_pPageInfoTemp->m_dwFlags & SWF_INT_CAPTIONPRESSED) != 0);
            m_pPageInfoTemp->m_dwFlags &= ~SWF_INT_CAPTIONPRESSED;
         }

         if (bRedraw)
         {
            CalcCaptionRect(m_pPageInfoTemp, &rect);
            CDC* pDC = GetDC();
            DrawPageCaption(pDC, m_pPageInfoTemp->m_lpszText, 
               m_pPageInfoTemp->m_dwFlags, &rect);
            ReleaseDC(pDC);
         }
      }
   }
}

void CStackWnd::OnLButtonDown(UINT nFlags, CPoint point) 
{
   BeginCaptionPress(point);
}

void CStackWnd::OnLButtonDblClk(UINT nFlags, CPoint point) 
{
   BeginCaptionPress(point);
}

void CStackWnd::OnLButtonUp(UINT nFlags, CPoint point) 
{
	if (m_bPanning)
      EndPanning();
   else
   {
      if (m_pPageInfoTemp)
      {
         BOOL bAction = FALSE, bRedraw = FALSE;
         CRect rect;

         if (CaptionHitTest(point) == m_pPageInfoTemp)
         {
            bAction = TRUE;
            bRedraw = ((m_pPageInfoTemp->m_dwFlags & SWF_INT_CAPTIONPRESSED) != 0);
            m_pPageInfoTemp->m_dwFlags &= ~SWF_INT_CAPTIONPRESSED;
         }

         if (bRedraw)
         {
            CalcCaptionRect(m_pPageInfoTemp, &rect);
            CDC* pDC = GetDC();
            DrawPageCaption(pDC, m_pPageInfoTemp->m_lpszText, 
               m_pPageInfoTemp->m_dwFlags, &rect);
            ReleaseDC(pDC);
         }

         ReleaseCapture();

         if (bAction)
         {
            SetPageState(m_pPageInfoTemp, TRUE, 
               (m_pPageInfoTemp->m_dwFlags & SWF_CONTRACTED) ? FALSE : TRUE);
         }

         m_pPageInfoTemp = NULL;
      }
   }
}

void CStackWnd::OnWindowPosChanged(WINDOWPOS FAR* lpwndpos) 
{
	CWnd::OnWindowPosChanged(lpwndpos);
	
   // only if window position relative to parent 
   // coordinates has changed
   if (m_ptWnd.x != lpwndpos->x || m_ptWnd.y != lpwndpos->y)
   {
      // redraw each page with all child controls
      CPageInfo* pInfo = m_pPageInfo;
      while (pInfo != NULL)
      {
         // redraw page
	      pInfo->m_pDlg->RedrawWindow(NULL, NULL, RDW_INVALIDATE | RDW_ERASE | 
            RDW_ALLCHILDREN | RDW_UPDATENOW);

         // retrieve next page
         pInfo = pInfo->m_pNext;
      }

      // save new window position
      m_ptWnd.x = lpwndpos->x;
      m_ptWnd.y = lpwndpos->y;
   }
}
